import random
from qsarmodelingpy.cross_validation_class import CrossValidation
import numpy as np
from sklearn.preprocessing import scale
from sklearn import linear_model


class YRandomization(object):
    def __init__(self, X, y, nLV, nrd=50):
        self.X = X
        self.y = y
        self.nLV = nLV
        self.nrd = nrd
        self.Q2 = []
        self.R2 = []
        self.R = []
        self.RMSECV = []
        for _ in range(nrd):
            yrd = y[random.sample(range(len(y)), len(y))]
            cv = CrossValidation(X, yrd, nLV)
            self.Q2.append(cv.Q2()[nLV-1])
            self.R2.append(cv.R2()[nLV-1])
            self.RMSECV.append(cv.RMSECV()[nLV-1])
            r = np.dot(np.transpose(scale(y)), scale(yrd))/(len(y))
            self.R.append(r[0][0])
        cv = CrossValidation(X, y, nLV)
        self.Q2.append(cv.Q2()[nLV-1])
        self.R2.append(cv.R2()[nLV-1])
        self.RMSECV.append(cv.RMSECV()[nLV-1])
        r = np.dot(np.transpose(scale(y)), scale(y))/(len(y))
        self.R.append(r[0][0])

    def returnIntercept(self):
        reg = linear_model.LinearRegression()
        reg.fit(np.array(self.R).reshape(-1, 1),
                np.array(self.R2).reshape(-1, 1))
        return reg.intercept_

    def returnRegResultsR2(self):
        reg = linear_model.LinearRegression()
        reg.fit(np.array(self.R).reshape(-1, 1),
                np.array(self.R2).reshape(-1, 1))
        return reg.coef_, reg.intercept_

    def returnRegResultsQ2(self):
        reg = linear_model.LinearRegression()
        reg.fit(np.array(self.R).reshape(-1, 1),
                np.array(self.Q2).reshape(-1, 1))
        return reg.coef_, reg.intercept_



"""Select samples using Kennard-Stone algorithm.

    > The KENNARDSTONE method selects a subset of samples from x which provide uniform coverage over the data set and includes samples on the boundary of the data set. The method begins by finding the two samples which are farthest apart using geometric distance. To add another sample to the selection set the algorithm selects from the remaining samples that one which has the greatest separation distance from the selected samples. The separation distance of a candidate sample from the selected set is the distance from the candidate to its closest selected sample. This most separated sample is then added to the selection set and the process is repeated until the required number of samples, k, have been added to the selection set. In practice this produces a very uniformly distributed network of selected points over the data set and includes samples along the boundary of the dataset. The method performs efficiently because it calculates the inter-sample distances matrix only once.

    ([Source](http://wiki.eigenvector.com/index.php?title=Kennardstone))
"""
import numpy as np


def kennardstonealgorithm(x_variables, k):
    """Run the Kennard-Stone algorithm.

    Args:
        x_variables (dataset): data to select samples from
        k (int): number of samples to be selected

    Returns:
        list: The indexes of selected samples (training data).
        list: The indexes of remaining samples (test data).
    """
    x_variables = np.array(x_variables)
    original_x = x_variables
    distance_to_average = (
        (x_variables - np.tile(x_variables.mean(axis=0), (x_variables.shape[0], 1))) ** 2).sum(axis=1)
    max_distance_sample_number = np.where(
        distance_to_average == np.max(distance_to_average))
    max_distance_sample_number = max_distance_sample_number[0][0]
    selected_sample_numbers = list()
    selected_sample_numbers.append(max_distance_sample_number)
    remaining_sample_numbers = np.arange(0, x_variables.shape[0], 1)
    x_variables = np.delete(x_variables, selected_sample_numbers, 0)
    remaining_sample_numbers = np.delete(
        remaining_sample_numbers, selected_sample_numbers, 0)
    for iteration in range(1, k):
        selected_samples = original_x[selected_sample_numbers, :]
        min_distance_to_selected_samples = list()
        for min_distance_calculation_number in range(0, x_variables.shape[0]):
            distance_to_selected_samples = ((selected_samples - np.tile(x_variables[min_distance_calculation_number, :],
                                                                        (selected_samples.shape[0], 1))) ** 2).sum(axis=1)
            min_distance_to_selected_samples.append(
                np.min(distance_to_selected_samples))
        max_distance_sample_number = np.where(
            min_distance_to_selected_samples == np.max(min_distance_to_selected_samples))
        max_distance_sample_number = max_distance_sample_number[0][0]
        selected_sample_numbers.append(
            remaining_sample_numbers[max_distance_sample_number])
        x_variables = np.delete(x_variables, max_distance_sample_number, 0)
        remaining_sample_numbers = np.delete(
            remaining_sample_numbers, max_distance_sample_number, 0)

    return selected_sample_numbers, remaining_sample_numbers



"""Implements filtering routines for dimensionality reduction.
    """

import numpy as np
import pandas as pd
from sklearn.preprocessing import scale as autoscale
import logging
from typing import Union

import qsarmodelingpy.lj_cut as lj


def variance_cut(X, cut):
    """Select only columns of `X` with a minimum variance `cut`.

    Args:
        X (DataFrame): The dataframe to be cutted
        cut (float): The minimum variance allowed

    Returns:
        list[int]: A list of all selected indexes.
    """
    v = np.var(X, 0, ddof=1)
    indCut = [i for i in range(len(v)) if v[i] >= cut]
    return indCut


def autoscale1(X):
    m, n = np.shape(X)
    means = np.mean(X, 0)
    stds = np.std(X, 0, ddof=1)
    Xm = X-np.ones((m, 1))*means
    Xa = np.divide(Xm, np.ones((m, 1)))
    return Xa


def correlation_cut(X, y, cut):
    """Select only columns of `X` with a minimum correlation `cut` with the dependent variable `y`.

    Args:
        X (DataFrame): The dataframe to be cutted
        y (DataFrame): The vector of dependent variable
        cut (float): The minimum correlation allowed

    Returns:
        list[int]: A list of all selected indexes.
    """
    corr = abs(np.dot(np.transpose(autoscale(X)), autoscale(y))/(len(y)-1))
    indCut = [i for i in range(len(corr)) if corr[i] >= cut]
    return indCut


def autocorrelation_cut(X: pd.DataFrame, y, cut):
    """Perform the matrix filtering based on autocorrelation.

    Args:
        X (DataFrame): The matrix to be filtered
        y (DataFrame): The vector of dependent variable
        cut (float): The filtering amount. If `cut == 1`, no cutting will be performed; if `cut == 0`, an empty list will be returned.

    Returns:
        list[int]: A list of all selected indexes.

    ## Example of use: 
    Let `X` and `y` be the dataset (matrix and vector). Then to perform an autocorrelation cut of 50%:

    ```python
    filtered_columns = autocorrelation_cut(X, y, 0.5)
    filtered_X = X.loc[:,filtered_columns]
    ```

    """
    m, n = X.shape
    Xcorr = np.corrcoef(X.T)
    var_filtered = []
    for i in range(n):
        for j in range(i+1, n):
            corr = Xcorr[i, j]
            if corr > cut:
                corr_i = abs(
                    np.dot(np.transpose(autoscale(X[:, i])), autoscale(y))/(m-1))
                corr_j = abs(
                    np.dot(np.transpose(autoscale(X[:, j])), autoscale(y))/(m-1))
                if corr_i < corr_j:
                    if i not in var_filtered:
                        var_filtered.append(i)
                else:
                    if j not in var_filtered:
                        var_filtered.append(j)
    var = [i for i in range(n) if i not in var_filtered]
    return var


def filter_matrix(
    X: pd.DataFrame,
    y,
    lj_transform: bool = False,
    var_cut: float = 0,
    corr_cut: float = 0,
    auto_corrcut: float = 1,
) -> pd.DataFrame:
    """Perform Lennard-Jones data transformation, variance cut, correlation cut and autocorrelation cut.

    While others methods in this class returns a list of indexes, `qsarmodelingpy.filter.filter_matrix` returns the filtered matrix ([`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html)).

    Args:
        X (DataFrame): The matrix to be filtered.
        y (DataFrame): The dependent variable
        lj_transform (bool, optional): Wether or not to perform Lennard-Jones Data Tranformation (see `qsarmodelingpy.lj_cut`). Defaults to False.
        var_cut (float, optional): See `qsarmodelingpy.filter.variance_cut`. Defaults to 0.
        corr_cut (float, optional): See `qsarmodelingpy.filter.correlation_cut`. Defaults to 0.
        auto_corrcut (float, optional): See `qsarmodelingpy.filter.autocorrelation_cut`. Defaults to 1.

    Returns:
        DataFrame: The filtered matrix.
    """
    # Pre processing
    if lj_transform:
        X = lj.transform(X)

    logging.info("Dimensions of the original matrix")
    logging.info(X.shape)

    # Variance Cut
    indVar = variance_cut(X.values, var_cut)
    X = X.loc[:, X.columns[indVar]]

    logging.info("Dimensions of the matrix after variance cut")
    logging.info(X.shape)

    # Correlation Cut
    indCorr = correlation_cut(X.values, y, corr_cut)
    X = X.loc[:, X.columns[indCorr]]

    logging.info("Dimensions of the matrix after correlation cut")
    logging.info(X.shape)

    # Autocorrelation cut
    indAuto = autocorrelation_cut(X.values, y, auto_corrcut)
    X = X.loc[:, X.columns[indAuto]]

    logging.info("Dimensions of the matrix after auto correlation cut")
    logging.info(X.shape)

    return X



import pandas as pd
import logging

from qsarmodelingpy.ops import OPS
from qsarmodelingpy.cross_validation_class import CrossValidation
from qsarmodelingpy.filter import filter_matrix
from qsarmodelingpy.validate_yr_lno import validate

from qsarmodelingpy.Interfaces import ConfigOPSInterface
# Uncomment the following line to see all console logs (if not yet done in main.py).
# logging.basicConfig(level=logging.DEBUG)


def run(config: ConfigOPSInterface) -> bool:
    """Run Ordered Predictors Selection.

    Args:
        config (qsarmodelingpy.Interfaces.ConfigOPSInterface): The configuration dictionary for OPS

    Returns:
        bool: `True` if validation pass, `False` otherwise.
    """

    # Open configuration file in order to look for the matrices and the parameters to run
    # OPS and cross-validation
    xFile = config['XMatrix']
    yFile = config['yvector']
    df = pd.read_csv(xFile, index_col=0)
    y = pd.read_csv(yFile, header=None).values
    var_cut = float(config['varcut'])
    corr_cut = float(config['corrcut'])
    autocorr_cut = float(config['autocorrcut'])
    nLVOPS = None if int(config['latent_vars_ops']) == 0 else int(
        config['latent_vars_ops'])
    nLVModel = None if int(config['latent_vars_model']) == 0 else int(
        config['latent_vars_model'])
    opsWindow = int(config['ops_window'])
    opsIncrement = int(config['ops_increment'])
    percentage = int(config['vars_percentage'])
    nModels = int(config['models_to_save'])
    yr_crit = float(config['yrand'])
    lno_crit = float(config['lno'])
    out_matrix = config['output_matrix']
    out_cv = config['output_cv']
    out_models = config['output_models']
    autoscale = config['autoscale']

    dfRest = filter_matrix(
        df, y, config['lj_transform'], var_cut, corr_cut, autocorr_cut)

    # Run
    dfRest.to_csv(out_matrix)
    X = dfRest.values
    ops = OPS(X, y, nLVOPS, nLVModel, opsWindow,
              opsIncrement, percentage, nModels, autoscale)
    typeOPS = config['ops_type']
    if typeOPS == 's':
        ops.runOPS()
    elif typeOPS == 'f':
        ops.feedOPS()
    else:
        raise Exception(
            f"Invalid option '{typeOPS}' for OPS type. Use 's' for single run or 'f' for feedOPS.")
    ops.saveModels(out_models)
    var_sel = validate(
        X, y, ops.models["var_sel"], ops.models["Q2"], yr_cut=yr_crit, lno_cut=lno_crit)
    if var_sel:
        dfSel = dfRest.loc[:, dfRest.columns[var_sel]]
        dfSel.to_csv(out_matrix)
        cv = CrossValidation(dfSel.values, y)
        cv.saveParameters(out_cv)
        return True
    else:
        logging.error("y-randomization or LNO failed!")
        return False



# Importing libraries
import numpy as np
import pandas as pd
from qsarmodelingpy.external_validation import ExternalValidation
from qsarmodelingpy.cross_validation_class import CrossValidation
from qsarmodelingpy.kennardstonealgorithm import kennardstonealgorithm
import qsarmodelingpy.lj_cut as lj
import os
import argparse
import logging


def run(config) -> None:
    """Run External Validation with configuration given by `config`.

    `config` is (usually) a CSV containing all needed information for this function. Please, [see this template](https://github.com/hellmrf/QSARModelingPyCore/blob/master/examples/confExtVal.csv) and edit for your needs.
    Probably, in future, `config` will also accept a Dictionary that follows rules defined in `qsarmodelingpy.Interfaces`.

    Args:
        config (str, path, file-like, `io`): The configuration file (usually a CSV). [See this template](https://github.com/hellmrf/QSARModelingPyCore/blob/master/examples/confExtVal.csv).
    """
    # Open configuration file in order to look for the matrices and the parameters to run
    # external validation
    dfConf = pd.read_csv(config, header=None)
    directory = dfConf[1][0]
    Xfile = dfConf[1][1]
    yfile = dfConf[1][2]
    nLV = None if dfConf.isnull()[1][4] else int(dfConf[1][4])
    out_directory = dfConf[1][5]
    ext_val_file = dfConf[1][6]
    cv_file = dfConf[1][7]
    Xtrain_file = dfConf[1][8]
    ytrain_file = dfConf[1][9]
    Xtest_file = dfConf[1][10]
    ytest_file = dfConf[1][11]
    autoscale = dfConf[1][12].upper() == "YES"
    y = pd.read_csv(os.path.join(directory, yfile),
                    sep=';', header=None).values
    dfX = pd.read_csv(os.path.join(directory, Xfile), sep=';', index_col=0)
    dfX = lj.transform(dfX) if dfConf[1][13].upper() == "YES" else dfX
    X = dfX.values
    type_ext_val = int(dfConf[1][14])
    if type_ext_val == 1:  # manual selection
        test_set = dfConf[1][3]
        test = [int(i) - 1 for i in test_set.split(',')]
        train = [j for j in range(len(y)) if j not in test]
    elif type_ext_val == 2:  # Kennard-Stone
        size_test_set = int(dfConf[1][3])
        # parameter is the size of training set
        train, test = kennardstonealgorithm(dfX, len(dfX) - size_test_set)
    else:  # Random selection
        pass
    ext = ExternalValidation(X, y, nLV)
    ext.extVal(train, test, nLV)
    if ext.validateExtVal(train, test):
        logging.info("Passed")
    else:
        logging.info("Failed")
    ext.saveExtVal(train, test, out_directory + "/" + ext_val_file)
    cv = CrossValidation(X[train, :], y[train], nLVMax=nLV, scale=True)
    cv.saveParameters(os.path.join(out_directory, cv_file))
    dfXtrain = dfX.loc[dfX.index[train], dfX.columns]
    dfXtrain.to_csv(os.path.join(out_directory, Xtrain_file), sep=';')
    dfytrain = pd.DataFrame(y[train])
    dfytrain.to_csv(os.path.join(out_directory, ytrain_file),
                    sep=',', header=False)
    dfXtest = dfX.loc[dfX.index[test], dfX.columns]
    dfXtest.to_csv(os.path.join(out_directory, Xtest_file), sep=';')
    dfytest = pd.DataFrame(y[test])
    dfytest.to_csv(os.path.join(out_directory, ytest_file),
                   sep=',', header=False)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('--filename', '-f', required=True,
                        metavar='<filename>',
                        help='Config External Validation file.')
    args = parser.parse_args()
    filename = args.filename
    run(filename)



import sys
import numpy as np
import pandas as pd
import json
from qsarmodelingpy.cross_validation_class import CrossValidation
from qsarmodelingpy.yrandomization import YRandomization
from qsarmodelingpy.lno import LNO
import qsarmodelingpy.lj_cut as lj
from qsarmodelingpy.filter import variance_cut, correlation_cut
from qsarmodelingpy.external_validation import ExternalValidation
import os


def validate(X, y, pop, Q2, Q2_cut=0.5, yr_cut=0.3, lno_cut=0.1):
    # y-randomization
    lpass = []
    intercepts = []
    for i, var_sel in enumerate(pop):
        if Q2[i] > Q2_cut:
            XSel = X[:, var_sel]
            cv = CrossValidation(XSel, y)
            nLV = np.argmax(cv.Q2())+1
            yr = YRandomization(XSel, y, nLV, 50)
            intercepts.append(yr.returnIntercept())
            if yr.returnIntercept() < yr_cut:
                lpass.append(i)
    # leave-N-out
    lpass2 = []
    if lpass != []:
        for i in lpass:
            if Q2[i] > Q2_cut:
                XSel = X[:, pop[i]]
                m, _ = np.shape(X)
                cv = CrossValidation(XSel, y)
                nLV = np.argmax(cv.Q2())+1
                lno = LNO(XSel, y, nLV, int(m/4), 5)
                m = np.mean(lno.Q2, 1)
                std = max([abs(m[j]-m[0]) for j in range(len(m))])
                if std < lno_cut:
                    lpass2.append(i)
        if lpass2 != []:
            return lpass2
        else:
            return []
    else:
        return []


def validate_ext(X, y, test, pop, Q2):
    train = [j for j in range(len(y)) if j not in test]
    passed = validate(X, y, pop, Q2)
    lpass = []
    pop_pass = []
    for i in passed:
        var_sel = pop[i]
        XSel = X[:, var_sel]
        cv = CrossValidation(XSel, y)
        nLV = np.argmax(cv.Q2())+1
        ext = ExternalValidation(XSel, y, nLV)
        if ext.validateExtVal(train, test, nLV):
            lpass.append(i)
            pop_pass.append(var_sel)
    return pop_pass


if __name__ == '__main__':
    directory = sys.argv[1]
    df = pd.read_csv(os.path.join(
        directory, sys.argv[2]), sep=';', index_col=0)
    dfX = lj.transform(df)
    y = pd.read_csv(os.path.join(
        directory, sys.argv[3]), sep=';', header=None).values
    indVar = variance_cut(dfX.values, 0.1)
    dfVar = dfX.loc[:, dfX.columns[indVar]]
    print(dfVar.shape)
    indCorr = correlation_cut(dfVar.values, y, 0.3)
    dfCorr = dfVar.loc[:, dfVar.columns[indCorr]]
    print(dfCorr.shape)
    out_directory = sys.argv[4]
    X = dfCorr.values
    with open(directory+"/Popout.json") as pop_file:
        pop = json.load(pop_file)
    with open(directory+"/Q2out.json") as Q2_file:
        Q2 = json.load(Q2_file)
    var_sel = validate(X, y, pop, Q2, yr_cut=0.25, lno_cut=0.1)
    if var_sel != []:
        dfSel = dfCorr.loc[:, dfCorr.columns[var_sel]]
        dfSel.to_csv(out_directory+"/XSel.csv", sep=';')
        cv = CrossValidation(dfSel.values, y)
        cv.saveParameters(out_directory+"/parameters_cv.csv")
    else:
        logging.warn("y-randomization or LNO failed!")



"""Variable selection with Genetic Algorithm."""
import random
import json

from deap import base
from deap import creator
from deap import tools
from numpy import zeros, shape, argmax
import pandas as pd
from pandas import read_csv
from qsarmodelingpy.cross_validation_class import CrossValidation


def returnIndices(individual):
    return [i for i, v in enumerate(individual) if v == 1]


def checkLen(min, max):
    def decorator(func):
        def wrapper(*args, **kargs):
            offspring = func(*args, **kargs)
            for child in offspring:
                size = sum(child)
                indices = returnIndices(child)
                indicesZeros = [i for i in list(
                    range(len(child))) if i not in indices]
                while size > max:
                    child[indices[random.randint(0, len(indices)-1)]] = 0
                    size = sum(child)
                while size < min:
                    child[indicesZeros[random.randint(
                        0, len(indicesZeros)-1)]] = 1
                    size = sum(child)
            return offspring
        return wrapper
    return decorator


def initIndividual(icls, imin, imax, size):
    indices = random.sample(range(size), random.randint(imin, imax))
    l = zeros(size)
    l[indices] = 1
    l = [int(i) for i in l.tolist()]
    ind = icls(l)
    return ind


class Ga(object):
    def __init__(self, X: pd.DataFrame, y: pd.DataFrame, nLV: int = None, scale: bool = True, min_size: int = 5, max_size: int = 25, size_population: int = 200, mig_rate: float = 0.2,
                 cxpb: float = 0.5, mutpb: float = 0.2, ngen: int = 120):
        """Variable selection with Genetic Algorithm.

        Args:
            X (DataFrame): The matrix with descriptors.
            y (DataFrame, list, array): The dependent variable vector.
            nLV (int, optional): Number of Latent Variables. Defaults to None.
            scale (bool, optional): Defaults to True.
            min_size (int, optional): Minimum number of variables in the model. Defaults to 5.
            max_size (int, optional): Maximum number of variables in the model. Defaults to 25.
            size_population (int, optional): Size of the population in each generation. Defaults to 200.
            mig_rate (float, optional): Migration rate. Defaults to 0.2.
            cxpb (float, optional): Crossover rate. Defaults to 0.5.
            mutpb (float, optional): Mutation rate. Defaults to 0.2.
            ngen (int, optional): Number of generations. Defaults to 120.
        """
        self.X = X
        self.y = y
        self.nLV = nLV
        self.scale = scale
        self.min_size = min_size
        self.max_size = max_size
        self.size_population = size_population
        self.mig_rate = mig_rate
        self.cxpb = cxpb
        self.mutpb = mutpb
        self.ngen = ngen

    def evaluate(self, individual):
        """Do some hard computing on the individual"""
        indices = [i for i, v in enumerate(individual) if v == 1]
        Xev = self.X[:, indices]
        cv = CrossValidation(Xev, self.y, min(
            self.nLV, len(indices)), self.scale)
        Q2 = max(cv.Q2())
        #yr = YRandomization(X,y,argmax(cv.Q2())+1,50)
        return (Q2,)

    def run(self):

        creator.create("FitnessMax", base.Fitness, weights=(1.0,))
        creator.create("Individual", list, fitness=creator.FitnessMax)

        MIN_SIZE = self. min_size
        MAX_SIZE = self.max_size
        IND_SIZE = shape(self.X)[1]
        SIZE_POPULATION = self.size_population
        MIG_RATE = self.mig_rate  # migration rate
        X = self.X
        y = self.y
        nLV = self.nLV
        CXPB = self.cxpb
        MUTPB = self.mutpb
        NGEN = self.ngen

        toolbox = base.Toolbox()

        toolbox.register("individual", initIndividual, creator.Individual,
                         MIN_SIZE, MAX_SIZE, IND_SIZE)

        toolbox.register("population", tools.initRepeat,
                         list, toolbox.individual)

        toolbox.register("mate", tools.cxTwoPoint)
        toolbox.register("mutate", tools.mutFlipBit, indpb=0.2)
        toolbox.decorate("mate", checkLen(MIN_SIZE, MAX_SIZE))
        toolbox.decorate("mutate", checkLen(MIN_SIZE, MAX_SIZE))
        toolbox.register("select", tools.selTournament, tournsize=2)
        #toolbox.register("select", tools.selBest)
        toolbox.register("evaluate", self.evaluate)

        pop = toolbox.population(n=SIZE_POPULATION)

        # Evaluate the entire population
        fitnesses = map(toolbox.evaluate, pop)
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = fit

        for g in range(NGEN):
            print("Generation {} of {}".format(g+1, NGEN))
            # generates migrants to add to the population
            mig = toolbox.population(n=int(MIG_RATE*SIZE_POPULATION))
            fitnesses = map(toolbox.evaluate, mig)
            for ind, fit in zip(mig, fitnesses):
                ind.fitness.values = fit
            # Select the next generation individuals
            # Clone the selected individuals
            offspring = list(map(toolbox.clone, pop))
            # form the new population
            offspring += mig
            indices = random.sample(range(len(offspring)), len(offspring))
            offspring = [offspring[i] for i in indices]

            # Apply crossover and mutation on the offspring
            for child1, child2 in zip(offspring[::2], offspring[1::2]):
                if random.random() < CXPB:
                    toolbox.mate(child1, child2)
                    del child1.fitness.values
                    del child2.fitness.values

            for mutant in offspring:
                if random.random() < MUTPB:
                    toolbox.mutate(mutant)
                    del mutant.fitness.values

            # Evaluate the individuals with an invalid fitness
            invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
            fitnesses = map(toolbox.evaluate, invalid_ind)
            for ind, fit in zip(invalid_ind, fitnesses):
                ind.fitness.values = fit

            # The population is replaced by best individuals in the offspring
            pop = toolbox.select(offspring, len(pop))

        individuals = [ind for ind in pop]
        self.Q2 = [ind.fitness.values for ind in pop]
        self.pop_selected = [returnIndices(ind) for ind in individuals]

    def saveQ2(self, file):
        """Saves Q² output to `file`.

        Args:
            file (str path, file-like, `io`): The filename to save Q² output.
        """
        with open(file, "w") as Q2_file:
            Q2_file.write(json.dumps(self.Q2))

    def savePop(self, file):
        """Saves population to `file`.

        Args:
            file (str path, file-like, `io`): The filename to save population output.
        """
        with open(file, "w") as pop_file:
            pop_file.write(json.dumps(self.pop_selected))



import pandas as pd
import logging

from qsarmodelingpy.Interfaces import ConfigGAInterface
from qsarmodelingpy.ga import Ga
from qsarmodelingpy.cross_validation_class import CrossValidation
from qsarmodelingpy.filter import filter_matrix
import qsarmodelingpy.lj_cut as lj
from qsarmodelingpy.validate_yr_lno import validate


def run(config: ConfigGAInterface) -> bool:
    """Run Generic Algorithm.

    Args:
        config (qsarmodelingpy.Interfaces.ConfigGAInterface): The configuration dictionary for GA

    Returns:
        bool: `True` if validation pass, `False` otherwise.
    """
    xFile = config['XMatrix']
    yFile = config['yvector']
    df = pd.read_csv(xFile, index_col=0)
    y = pd.read_csv(yFile, header=None).values
    var_cut = float(config['varcut'])
    corr_cut = float(config['corrcut'])
    autocorr_cut = float(config['autocorrcut'])
    nLVModel = None if int(config['max_latent_model']) == 0 else int(
        config['max_latent_model'])
    min_size = int(config['min_vars_model'])
    max_size = int(config['max_vars_model'])
    size_population = int(config['population_size'])
    mig_rate = float(config['migration_rate'])
    cxpb = float(config['crossover_rate'])
    mutpb = float(config['mutation_rate'])
    ngen = int(config['generations'])
    yr_crit = float(config['yrand'])
    lno_crit = float(config['lno'])
    out_matrix = config['output_matrix']
    out_cv = config['output_cv']
    Q2_file = config['output_q2']
    var_sel_file = config['output_selected']
    autoscale = config['autoscale']

    dfRest = filter_matrix(
        df, y, config['lj_transform'], var_cut, corr_cut, autocorr_cut)

    # Run
    X = dfRest.values
    if nLVModel is None:
        nLVModel = int(dfRest.shape[0]/5)
    ga = Ga(X, y, nLVModel, autoscale, min_size, max_size,
            size_population, mig_rate, cxpb, mutpb, ngen)
    ga.run()
    ga.saveQ2(Q2_file)
    ga.savePop(var_sel_file)
    Q2 = ga.Q2
    Q2 = [Q2[i][0] for i, _ in enumerate(Q2)]
    var_sel = validate(X, y, ga.pop_selected, Q2,
                       yr_cut=yr_crit, lno_cut=lno_crit)
    if var_sel:
        dfSel = dfRest.loc[:, dfRest.columns[var_sel]]
        dfSel.to_csv(out_matrix)
        cv = CrossValidation(dfSel.values, y)
        cv.saveParameters(out_cv)
        logging.info("Done Genetic Algorithm")
        return True
    else:
        logging.error("y-randomization or Leave-N-Out failed!")
        return False



r"""Implements a data transformation for Lennard-Jones data. When some calculation point happens to be "inside" an atom, Lennard-Jones interaction goes to infinity. This module implement a workaround using a logarithm to encode values greater than a limit `cut`.
"""
import numpy as np
import math


def ljCut(value, cut):
    calValue = value/4.18
    if calValue >= cut:
        calValue = cut + math.log10(calValue-(cut-1))
    calValue = calValue*4.18
    return calValue


def transform(dfLj):
    m, n = dfLj.shape
    for i in range(m):
        for j in range(n):
            dfLj.values[i, j] = ljCut(dfLj.values[i, j], 30)
    return dfLj



from sklearn.model_selection import ShuffleSplit
from sklearn.cross_decomposition import PLSRegression
from qsarmodelingpy.cross_validation_class import CrossValidation
from qsarmodelingpy.calculate_parameters import ssy, calcPress, calcR, calcR2, calcMAE, calcRMSE
import numpy as np
import pandas as pd
import math
import sys
import json
import os
import logging


class ExternalValidation(object):
    def __init__(self, X, y, nLV=None):
        super(ExternalValidation, self).__init__()
        self.X = X
        self.y = y
        self.nLV = nLV

    def validateExtVal(self, train, test, nLV=None):
        X = self.X
        y = self.y
        nLV = self.nLV
        if nLV == None:
            cv = CrossValidation(X[train, :], y[train])
            nLV = np.argmax(cv.Q2())+1
        pls = PLSRegression(n_components=nLV)
        pls.fit(X[train, :], y[train])
        yp = pls.predict(X[test])
        Q2F1 = calcR2(y[test], yp, np.mean(y[train]))
        Q2F2 = calcR2(y[test], yp)
        MAE = calcMAE(y[test], yp)
        residuals = np.reshape(yp-y[test], len(test))
        sd = np.std(abs(residuals), ddof=1)
        higher_residuals = np.argsort(-abs(residuals)
                                      )[:math.ceil(0.05*len(test))]
        higher_residuals = [test[i] for i in higher_residuals]
        new_test = [i for i in test if i not in higher_residuals]
        yp = pls.predict(X[new_test])
        Q2F195 = calcR2(y[new_test], yp, np.mean(y[train]))
        Q2F295 = calcR2(y[new_test], yp)
        MAE95 = calcMAE(y[new_test], yp)
        residuals95 = np.reshape(abs(yp-y[new_test]), len(new_test))
        sd95 = np.std(residuals95, ddof=1)
        tr = max(y[train]) - min(y[train])
        NE = [i for i in residuals if i < 0]
        PE = [i for i in residuals if i > 0]
        nNE = len(NE)
        nPE = len(PE)
        mNE = abs(np.mean(NE))
        mPE = np.mean(PE)
        if (nNE == 0) or (nPE == 0) or (nNE/nPE > 5) or (nPE/nNE > 5):
            return False
        if (MAE95 <= 0.1*tr) and (MAE95 + 3*sd95 <= 0.2*tr):
            if Q2F1 > 0.5 and Q2F2 > 0.5:
                return True
            else:
                return False
        else:
            return False

    def searchValidExtVal(self, nLV=None, n_test=10, n_splits=100):
        X = self.X
        y = self.y
        nLV = self.nLV
        if nLV == None:
            cv = CrossValidation(X, y)
            nLV = np.argmax(cv.Q2())+1
        N_SPLITS = n_splits
        TEST_SIZE = n_test/len(y)
        ss = ShuffleSplit(n_splits=N_SPLITS, test_size=TEST_SIZE)
        pls = PLSRegression(n_components=nLV)
        R2 = np.zeros(0)
        Q2 = np.zeros(0)
        i = 0
        tests_sets = []
        for train, test in ss.split(X):
            # teste com conjunto teste fixo para ver se parametros estão sendo calculados corretamente
            # test = [23, 47, 6, 25, 40, 13, 48, 30, 43, 19]
            # train = [i for i in range(len(y)) if i not in test]
            cv = CrossValidation(X[train, :], y[train])
            nLV = np.argmax(cv.Q2())+1
            pls = PLSRegression(n_components=nLV)
            pls.fit(X[train, :], y[train])
            yp = pls.predict(X[test])
            Q2F1 = calcR2(y[test], yp, np.mean(y[train]))
            Q2F2 = calcR2(y[test], yp)
            MAE = calcMAE(y[test], yp)
            residuals = np.reshape(yp-y[test], len(test))
            sd = np.std(abs(residuals), ddof=1)
            higher_residuals = np.argsort(-abs(residuals)
                                          )[:math.ceil(0.05*n_test)]
            higher_residuals = [test[i] for i in higher_residuals]
            new_test = [i for i in test if i not in higher_residuals]
            yp = pls.predict(X[new_test])
            Q2F195 = calcR2(y[new_test], yp, np.mean(y[train]))
            Q2F295 = calcR2(y[new_test], yp)
            MAE95 = calcMAE(y[new_test], yp)
            residuals95 = np.reshape(abs(yp-y[new_test]), len(new_test))
            sd95 = np.std(residuals95, ddof=1)
            tr = max(y[train]) - min(y[train])
            NE = [i for i in residuals if i < 0]
            PE = [i for i in residuals if i > 0]
            nNE = len(NE)
            nPE = len(PE)
            mNE = abs(np.mean(NE))
            mPE = np.mean(PE)
            if (nNE == 0) or (nPE == 0) or (nNE/nPE > 5) or (nPE/nNE > 5):
                logging.debug("Warning")
            if (MAE95 <= 0.1*tr) and (MAE95 + 3*sd95 <= 0.2*tr):
                logging.debug("good")
                tests_sets.append(test)
                R2 = np.r_[R2, Q2F2]
                Q2 = np.r_[Q2, max(cv.Q2())]
            else:
                if (MAE95 > 0.15*tr) or (MAE95 + 3*sd95 > 0.25*tr):
                    print("bad")
                else:
                    print("moderate")

        ind = np.argsort(-R2)
        self.R2 = R2[ind]
        self.Q2 = Q2[ind]
        self.tests_sets = [tests_sets[i] for i in ind]

    def extVal(self, train, test, nLV=None):
        Xtest = self.X[test]
        Xtrain = self.X[train]
        ytest = self.y[test]
        ytrain = self.y[train]
        nLV = self.nLV
        if nLV == None:
            cv = CrossValidation(Xtrain, ytrain)
            nLV = np.argmax(cv.Q2())+1
        pls = PLSRegression(n_components=nLV)
        pls.fit(Xtrain, ytrain)
        ypred = pls.predict(Xtest)
        self.ypred = ypred
        Q2F1 = calcR2(ytest, ypred, np.mean(ytrain))
        Q2F2 = calcR2(ytest, ypred)
        Q2F3 = 1-((ytest-ypred).T.dot(ytest-ypred)/len(ytest)) / \
            ((ytrain-ytrain.mean()).T.dot(ytrain-ytrain.mean())/len(ytrain))
        residuals = np.reshape(ypred-ytest, len(test))
        higher_residuals = np.argsort(-abs(residuals)
                                      )[:math.ceil(0.05*len(test))]
        higher_residuals = [test[i] for i in higher_residuals]
        new_test = [i for i in test if i not in higher_residuals]
        yp = pls.predict(self.X[new_test])
        Q2F195 = calcR2(self.y[new_test], yp, np.mean(ytrain))
        Q2F295 = calcR2(self.y[new_test], yp)
        Q2F395 = 1-((self.y[new_test]-yp).T.dot(self.y[new_test]-yp)/len(yp)) / \
            ((ytrain-ytrain.mean()).T.dot(ytrain-ytrain.mean())/len(ytrain))
        MAE95 = calcMAE(self.y[new_test], yp)
        k = sum(ytest*ypred)/sum(ypred**2)
        k1 = sum(ytest*ypred)/sum(ytest**2)
        yr0 = k*ypred
        y1r0 = k1*ytest
        R02 = calcR2(ytest, yr0)
        R102 = calcR2(ypred, y1r0)
        dif = abs(R02-R102)
        AREpred = sum(abs(ytest-ypred)/ytest)*100/len(ytest)
        RMSEP = calcRMSE(ytest, ypred)
        scaledy_test = (ytest-min(ytest))/(max(ytest)-min(ytest))
        scaledyp = (ypred-min(ytest))/(max(ytest)-min(ytest))
        scaled_k = sum(scaledy_test*scaledyp)/sum(scaledyp**2)
        scaled_k1 = sum(scaledy_test*scaledyp)/sum(scaledy_test**2)
        scaled_yr0 = scaled_k*scaledyp
        scaled_y1r0 = scaled_k1*scaledy_test
        scaled_R02 = calcR2(scaledy_test, scaled_yr0)
        scaled_R102 = calcR2(scaledyp, scaled_y1r0)
        r2 = (calcR(np.reshape(scaledy_test, len(scaledy_test)),
                    np.reshape(scaledyp, len(scaledy_test))))**2
        rm2 = r2*(1-math.sqrt(r2-scaled_R02))
        rm12 = r2*(1-math.sqrt(r2-scaled_R102))
        avgrm = (rm2+rm12)/2
        deltarm = abs(rm2-rm12)
        MAE = calcMAE(ytest, ypred)
        test_set = str([i+1 for i in test])
        dfPred = pd.DataFrame(index=["Q2F1", "Q2F2", "Q2F3", "Q2F195", "Q2F295", "Q2F395",
                                     "k", "k'", "R0",
                                     "R0'", "dif", "AREpred", "RMSEP", "avgrm",
                                     "deltarm", "MAE", "MAE95", "test_set", "nLV"],
                              data=np.array([Q2F1[0], Q2F2[0], Q2F3[0][0], Q2F195[0], Q2F295[0], Q2F395[0][0],
                                             k[0], k1[0], R02[0], R102[0], dif[0], AREpred[0], RMSEP, avgrm,
                                             deltarm, MAE, MAE95, test_set, nLV]))
        return dfPred

    def saveExtVal(self, train, test, fileName):
        dfPred = self.extVal(train, test, self.nLV)
        dfPred.to_csv(fileName, sep=',', header=False)

    def runValidExtVal(self, directory, n_models=1):
        print("Entrou na funcao de testar validacoes")
        print(len(self.tests_sets))
        for i in range(min(n_models, len(self.tests_sets))):
            test = self.tests_sets[i]
            train = [j for j in range(len(self.y)) if j not in test]
            dfPred = self.extVal(train, test)
            dfPred.to_csv(directory+"test"+str(i+1) +
                          ".csv", sep=',', header=False)

    def searchValidExtVal2(self, directory, nLV=None, n_test=10, n_splits=100):
        X = self.X
        y = self.y
        nLV = self.nLV
        if nLV == None:
            cv = CrossValidation(X, y)
            nLV = np.argmax(cv.Q2())+1
        N_SPLITS = n_splits
        TEST_SIZE = n_test/len(y)
        ss = ShuffleSplit(n_splits=N_SPLITS, test_size=TEST_SIZE)
        pls = PLSRegression(n_components=nLV)
        i = 0
        dfExt = pd.DataFrame()
        dfCv = pd.DataFrame()
        for train, test in ss.split(X):
            if self.validateExtVal(train, test, nLV):
                logging.debug("Achou um conjunto teste valido")
                dfExt[i] = self.extVal(train, test, nLV).iloc[:, 0]
                cv = CrossValidation(
                    X[train, :], y[train], nLVMax=nLV, scale=True)
                dfCv[i] = cv.returnParameters(nLV).iloc[:, 0]
                i = i + 1
        return dfExt, dfCv


#######################################
#######################################
#######################################
#######################################
#######################################
#######################################

# from qsarmodelingpy.cross_validation_class import CrossValidation
# from qsarmodelingpy.external_validation import ExternalValidation
# import pandas as pd
# import numpy as np
# import sys


# sys.exit(0)


# if __name__ == '__main__':
#     dfConf = pd.read_csv("confExtVal.csv", header=None)
#     directory = dfConf[1][0]
#     Xfile = dfConf[1][1]
#     yfile = dfConf[1][2]
#     test_set = dfConf[1][3]
#     nLV = None if dfConf.isnull()[1][4] else int(dfConf[1][4])
#     out_directory = dfConf[1][5]
#     ext_val_file = dfConf[1][6]
#     cv_file = dfConf[1][7]
#     Xtrain_file = dfConf[1][8]
#     ytrain_file = dfConf[1][9]
#     Xtest_file = dfConf[1][10]
#     ytest_file = dfConf[1][11]
#     autoscale = dfConf[1][12].upper() == "YES"
#     dfX = pd.read_csv(directory+"/"+Xfile, sep=';', index_col=0)
#     y = pd.read_csv(directory+"/"+yfile, sep=';', header=None).values
#     X = dfX.values
#     ext = ExternalValidation(X, y, nLV)
#     test = [int(i)-1 for i in test_set.split(',')]
#     train = [j for j in range(len(y)) if j not in test]
#     ext.extVal(train, test)
#     ext.saveExtVal(train, test, out_directory+"/"+ext_val_file)
#     cv = CrossValidation(X[train, :], y[train], nLVMax=nLV, scale=True)
#     cv.saveParameters(out_directory+"/"+cv_file)
#     # dfXtrain = pd.DataFrame(X[train,:],columns=dfX.columns)
#     dfXtrain = dfX.loc[dfX.index[train], dfX.columns]
#     dfXtrain.to_csv(out_directory+"/"+Xtrain_file, sep=';')
#     dfytrain = pd.DataFrame(y[train])
#     dfytrain.to_csv(out_directory+"/"+ytrain_file, sep=',', header=False)
#     # dfXtest = pd.DataFrame(X[test,:],columns=dfX.columns)
#     dfXtest = dfX.loc[dfX.index[test], dfX.columns]
#     dfXtest.to_csv(out_directory+"/"+Xtest_file, sep=';')
#     dfytest = pd.DataFrame(y[test])
#     dfytest.to_csv(out_directory+"/"+ytest_file, sep=',', header=False)



"""Define algorithms for some statistics, such as
    [SSY](https://en.wikipedia.org/wiki/Total_sum_of_squares),
    [PRESS](https://en.wikipedia.org/wiki/PRESS_statistic),
    [R²](https://en.wikipedia.org/wiki/Coefficient_of_determination),
    [MAE](https://en.wikipedia.org/wiki/Mean_absolute_error),
    [RMSE](https://en.wikipedia.org/wiki/Root-mean-square_deviation),
    `R`…
    """
import numpy as np
from math import sqrt
from sklearn.preprocessing import scale


def ssy(y, mean_y=None) -> float:
    r"""Calculates [Total Sum of Squares](https://en.wikipedia.org/wiki/Total_sum_of_squares) of the dependent variable.

    \[ SSY = \sum\limits_{i} (y_i - \bar y)^2 \]

    Args:
        y (DataFrame, list, array): The vector to calculate SSY.
        mean_y (float, optional): Mean of `y`. If `None`, it'll be calculated. Defaults to None.

    Returns:
        float: SSY
    """
    if mean_y is None:
        mean_y = np.mean(y)
    ssy = sum((y-mean_y*np.ones(np.shape(y)))**2)
    return ssy


def calcPress(yreal, ypred) -> float:
    r"""Calculates [predicted residual error sum of squares](https://en.wikipedia.org/wiki/PRESS_statistic) of `ypred` regarding `yreal`.

    \[ PRESS = \sum\limits_{i} (y_i - \hat y_{i})^2 \]

    Args:
        yreal (DataFrame,list,array): The real data.
        ypred (DataFrame,list,array): The predicted data.

    Returns:
        float: PRESS
    """
    press = sum((yreal-ypred)**2)
    return press


def calcR2(yreal, ypred, mean_y=None) -> float:
    r"""Calculates the [coefficient of determination](https://en.wikipedia.org/wiki/Coefficient_of_determination) ( \( R^2 \) ).

    \[ R^2 = 1-\dfrac{PRESS}{SSY} \]

    Args:
        yreal (DataFrame,list,array): The real data.
        ypred (DataFrame,list,array): The predicted data.
        mean_y (float, optional): Mean of `yreal`. If `None`, it'll be calculated. Defaults to None.

    Returns:
        float: \( R^2 \)

    ## See also:
    * `qsarmodelingpy.calculate_parameters.calcPress`
    * `qsarmodelingpy.calculate_parameters.ssy`
    """
    ssy1 = ssy(yreal, mean_y)
    R2 = 1-(calcPress(yreal, ypred)/ssy1)
    return R2


def calcMAE(yreal, ypred) -> float:
    r"""Calculates the [Mean Absolute Error](https://en.wikipedia.org/wiki/Mean_absolute_error).

    \[ MAE = \dfrac{\sum\limits_{i=1}^n \left\vert y_i - \hat y_i \right\vert}{n} \]

    where \(y\) is the predicted value, \(\hat y\) is the true value and \(n\) is the number of data points.

    Args:
        yreal (DataFrame,list,array): The real data.
        ypred (DataFrame,list,array): The predicted data.

    Returns:
        float: MAE
    """
    MAE = 1/len(yreal)*np.sum(abs(yreal-ypred))
    return MAE


def calcRMSE(yreal, ypred) -> float:
    r"""Calculates the [Root Mean Square Error](https://en.wikipedia.org/wiki/Root-mean-square_deviation).

    \[ RMSE = \sqrt{\dfrac{PRESS}{n}} \]

    Args:
        yreal (DataFrame,list,array): The real data.
        ypred (DataFrame,list,array): The predicted data.

    Returns:
        float: RMSE

    ## See also:
    * `qsarmodelingpy.calculate_parameters.calcPress`
    """
    RMSE = sqrt(calcPress(yreal, ypred)/len(yreal))
    return RMSE


# TODO: expand documentation
def calcR(yreal, ypred) -> float:
    r"""Calculates \( R \).

    Args:
        yreal (DataFrame,list,array): The real data.
        ypred (DataFrame,list,array): The predicted data.

    Returns:
        float: R
    """
    r = np.dot(scale(yreal), scale(ypred))/(len(yreal))
    return r



from qsarmodelingpy.leaveNout import LeaveNout
import numpy as np
from sklearn.cross_decomposition import PLSRegression
import qsarmodelingpy.calculate_parameters as cp


class LNO(object):

    def __init__(self, X, y, nLV, n=None, nrepeat=5):
        self.X = X
        self.y = y
        self.nLV = nLV if nLV is not None else len(y)/5
        self.n = n if n is not None else int(len(y)/4)
        self.nrepeat = nrepeat
        ycv = np.zeros((len(y), 1))
        self.Q2 = np.zeros((self.n, nrepeat))
        i = 0
        pls = PLSRegression(n_components=nLV)
        n_samples = np.shape(X)[0]
        for i in range(self.n):
            lno = LeaveNout(i+1, n_repeats=self.nrepeat)
            nsplits = int(n_samples/(i+1))
            if n_samples % (i+1) != 0:
                nsplits += 1
            j = 0
            k = 0
            for train, test in lno.split(X):
                pls.fit(X[train, :], y[train])
                ycv[test, :] = pls.predict(X[test, :])
                j = (j+1) % nsplits
                if j == 0:
                    self.Q2[i, k] = cp.calcR2(self.y, ycv)
                    k = k + 1



import random
import numpy as np


class LeaveNout(object):
    def __init__(self, n, n_repeats):
        r"""[Leave-N-Out](https://link.springer.com/referenceworkentry/10.1007%2F978-0-387-30164-8_469) Cross-Validation

        Args:
            n (int): \( n \) to leave out.
            n_repeats (int): Repetitions.
        """

        self.n = n
        self.n_repeats = n_repeats

    def split(self, X):
        n_samples = np.shape(X)[0]
        n = self.n
        n_splits = int(n_samples/n)
        result = []
        for i in range(self.n_repeats):
            indices = random.sample(range(n_samples), n_samples)
            for j in range(n_splits):
                testSet = indices[j*n:n*(j+1)]
                trainingSet = [k for k in indices if k not in testSet]
                result.append((trainingSet, testSet))
            n_remaining = n_samples % n
            if n_remaining != 0:
                testSet = indices[-n_remaining:]
                trainingSet = [k for k in indices if k not in testSet]
                result.append((trainingSet, testSet))
        return result



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qsarmodelingpy",
    version="0.1.1",
    author="Martins, J. P. A; Reis Filho, H. M.",
    author_email="jpam@qui.ufmg.br,helitonmrf@ufmg.br",
    description="A library for building and validating QSAR models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hellmrf/QSARModelingPyCore",
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'scikit-learn',
        'deap',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    package_data={
        '': ['Fcritico.xlsx'],
    },
)



import numpy as np
from qsarmodelingpy.cross_validation_class import CrossValidation
from qsarmodelingpy.plsbdg import PLSBidiag
from sklearn.preprocessing import scale as autoscale
import operator
import json
import logging
import pandas as pd


class OPS(object):
    def __init__(self, X: pd.DataFrame, y: pd.DataFrame, nLV: int = None, nLVModel: int = None, window: int = 2, increment: int = 1, percentage=100,
                 nModels=100, scale=True):
        super(OPS, self).__init__()
        self.X = autoscale(X) if scale else X
        self.y = autoscale(y) if scale else y
        self.nLV = min(np.shape(X)) if nLV is None else nLV
        self.nLVModel = int(np.shape(X)[0]/5) if nLVModel is None else nLVModel
        if window >= 2:
            self.window = window
        else:
            self.window = 2
            logging.warning(
                "For OPS, window parameter must be at least 2. Continuing with this value.")
        self.increment = increment
        self.percentage = percentage
        self.nModels = nModels
        self.vars = np.array(range(np.shape(X)[1]))

    def vectors(self, X, y, nLV):
        correlogram = abs(
            np.dot(np.transpose(autoscale(X)), autoscale(y))/(len(y)-1))
        pls = PLSBidiag(nLV)
        pls.fit(X, y)
        reg = abs(pls.B)
        vec = abs(np.append(correlogram, reg, axis=1))
        norms = np.linalg.norm(vec, ord=np.inf, axis=0)
        for i in range(len(norms)):
            vec[:, i] = vec[:, i]/norms[i]
        for i in range(1, len(norms)):
            vec = np.c_[vec, np.multiply(vec[:, 0], vec[:, i])]
        return vec

    def runOPS(self, X=None, y=None):
        if X is None:
            X = self.X
        if y is None:
            y = self.y
        nLVModel = self.nLVModel
        m, n = np.shape(X)
        vec = self.vectors(X, y, self.nLV)
        nVec = 2*self.nLV+1
        maxVar = self.percentage*n/100
        models = {}
        models["Q2"] = []
        models["var_sel"] = []
        Q2 = np.zeros(0)
        var_sel = []
        for i in range(nVec):
            logging.info("Running vector {} of {}".format(i+1, nVec))
            # ordenar em ordem decrescente, por isso o -vec
            ind = np.argsort(-vec[:, i])
            Xor = X[:, ind]
            nVar = self.window
            while nVar <= maxVar:
                Xev = Xor[:, 0:nVar]
                cv = CrossValidation(Xev, y, nLVModel)
                # models["Q2"].append(max(cv.Q2()))
                # models["var_sel"].append(ind[0:nVar])
                Q2 = np.append(Q2, max(cv.Q2()))
                # var_sel.append(self.vars[ind[0:nVar]].tolist())
                var_sel.append([int(self.vars[ind[i]]) for i in range(nVar)])
                nVar += self.increment
            indSort = np.argsort(-Q2).tolist()
            Q2 = Q2[indSort]
            var_sel = [var_sel[k] for k in indSort]
            Q2 = Q2[0:min(self.nModels, len(Q2))]
            var_sel = var_sel[0:min(self.nModels, len(var_sel))]
        models["Q2"] = Q2.tolist()
        models["var_sel"] = var_sel
        self.models = models

    def feedOPS(self):
        print("First run of OPS")
        self.runOPS()
        var_sel1 = []
        X = self.X
        y = self.y
        Q2 = 0
        # and np.sort(var_sel1) != np.sort(self.models["var_sel"][0]):
        while Q2 < self.models["Q2"][0]:
            var_sel1 = self.models["var_sel"][0]
            Q2 = self.models["Q2"][0]
            print("{} variables were selected in the previous step of OPS run".format(
                len(var_sel1)))
            X = self.X[:, var_sel1]
            self.vars = var_sel1
            # with less than 100 variables we could run OPS with 100 % of variables
            if np.shape(X)[1] < 100:
                self.percentage = 100
            self.nLV = min(np.shape(X))
            self.runOPS(X, y)

    def saveModels(self, file):
        with open(file, "w") as models_file:
            models_file.write(json.dumps(self.models))



""" Cross-Validation routines.

From [Wikipedia](https://en.wikipedia.org/wiki/Cross-validation_(statistics)):

> Cross-validation is any of various similar model validation techniques for assessing how the results of a statistical analysis will generalize to an independent data set.
It is mainly used in settings where the goal is prediction, and one wants to estimate how accurately a predictive model will perform in practice.
In a prediction problem, a model is usually given a dataset of known data on which training is run (training dataset), and a dataset of unknown data (or first seen data) against which the model is tested (called the validation dataset or testing set).
The goal of cross-validation is to test the model's ability to predict new data that was not used in estimating it, in order to flag problems like overfitting or selection bias and to give an insight on how the model will generalize to an independent dataset (i.e., an unknown dataset, for instance from a real problem). 
"""
from sklearn.model_selection import LeaveOneOut
from sklearn.cross_decomposition import PLSRegression
import numpy as np
import pandas as pd
from math import sqrt
from sklearn.preprocessing import scale
from qsarmodelingpy.calculate_parameters import calcPress, calcR2, calcRMSE, calcR
from qsarmodelingpy.plsbdg import PLSBidiag


class CrossValidation(object):

    def __init__(self, X: pd.DataFrame, y, nLVMax: int = None, scale: bool = True) -> None:
        """ Perform the Cross Validation *via* Leave-One-Out.

        Args:
            X (DataFrame): The matrix with descriptors.
            y (DataFrame, list, array): The dependent variable vector.
            nLVMax (int, optional): Maximum number of Latent Variables. Defaults to None.
            scale (bool, optional): Defaults to True.

        ## See also:
        * `qsarmodelingpy.leaveNout`
        """
        self.X = X
        self.y = y
        if nLVMax == None:
            self.nLVMax = int(len(y)/5)
        else:
            self.nLVMax = nLVMax
        nLVMax = self.nLVMax
        # Cross validation
        ycv = np.zeros((len(y), nLVMax))
        loo = LeaveOneOut()
        for train, test in loo.split(X):
            pls = PLSBidiag(n_components=nLVMax)
            pls.fit(X[train, :], y[train])
            ycv[test, :] = pls.predict(X[test])

        # Calibration
        ycal = np.zeros((len(y), nLVMax))
        pls = PLSBidiag(n_components=nLVMax)
        pls.fit(X, y)
        ycal = pls.predict(X)

        self.ycv = ycv
        self.ycal = ycal

    def press(self):
        return [calcPress(self.y[:, 0], self.ycal[:, i]) for i in range(self.nLVMax)]

    def R2(self):
        return [calcR2(self.y[:, 0], self.ycal[:, i]) for i in range(self.nLVMax)]

    def RMSEC(self):
        return [calcRMSE(self.y[:, 0], self.ycal[:, i]) for i in range(self.nLVMax)]

    def rcal(self):
        return [calcR(self.y[:, 0], self.ycal[:, i]) for i in range(self.nLVMax)]

    def presscv(self):
        return [calcPress(self.y[:, 0], self.ycv[:, i]) for i in range(self.nLVMax)]

    def Q2(self):
        return [calcR2(self.y[:, 0], self.ycv[:, i]) for i in range(self.nLVMax)]

    def RMSECV(self):
        return [calcRMSE(self.y[:, 0], self.ycv[:, i]) for i in range(self.nLVMax)]

    def rcv(self):
        return [calcR(self.y[:, 0], self.ycv[:, i]) for i in range(self.nLVMax)]

    def returnParameters(self, nLV=None):
        if nLV == None:
            nLV = np.argmax(self.Q2())+1
        press_value = calcPress(self.y[:, 0], self.ycal[:, nLV-1])
        R2_value = calcR2(self.y[:, 0], self.ycal[:, nLV-1])
        RMSEC_value = calcRMSE(self.y[:, 0], self.ycal[:, nLV-1])
        rcal_value = calcR(self.y[:, 0], self.ycal[:, nLV-1])
        presscv_value = calcPress(self.y[:, 0], self.ycv[:, nLV-1])
        Q2_value = calcR2(self.y[:, 0], self.ycv[:, nLV-1])
        RMSECV_value = calcRMSE(self.y[:, 0], self.ycv[:, nLV-1])
        rcv_value = calcR(self.y[:, 0], self.ycv[:, nLV-1])
        scaledy = (self.y[:, 0]-min(self.y[:, 0])) / \
            (max(self.y[:, 0])-min(self.y[:, 0]))
        scaledycal = (self.ycal[:, nLV-1]-min(self.y[:, 0])) / \
            (max(self.y[:, 0])-min(self.y[:, 0]))
        scaled_k = sum(scaledy*scaledycal)/sum(scaledycal**2)
        scaled_k1 = sum(scaledy*scaledycal)/sum(scaledy**2)
        scaled_yr0 = scaled_k*scaledycal
        scaled_y1r0 = scaled_k1*scaledy
        scaled_R02 = calcR2(scaledy, scaled_yr0)
        scaled_R102 = calcR2(scaledycal, scaled_y1r0)
        r2cal = (calcR(np.reshape(scaledy, len(scaledy)),
                       np.reshape(scaledycal, len(scaledy))))**2
        rm2 = r2cal*(1-sqrt(abs(r2cal-scaled_R02)))
        rm12 = r2cal*(1-sqrt(abs(r2cal-scaled_R102)))
        avgrmcal = (rm2+rm12)/2
        deltarmcal = abs(rm2-rm12)
        scaledycv = (self.ycv[:, nLV-1]-min(self.y[:, 0])) / \
            (max(self.y[:, 0])-min(self.y[:, 0]))
        scaled_k = sum(scaledy*scaledycv)/sum(scaledycv**2)
        scaled_k1 = sum(scaledy*scaledycv)/sum(scaledy**2)
        scaled_yr0 = scaled_k*scaledycv
        scaled_y1r0 = scaled_k1*scaledy
        scaled_R02 = calcR2(scaledy, scaled_yr0)
        scaled_R102 = calcR2(scaledycv, scaled_y1r0)
        r2cv = (calcR(np.reshape(scaledy, len(scaledy)),
                      np.reshape(scaledycv, len(scaledy))))**2
        rm2 = r2cv*(1-sqrt(abs(r2cv-scaled_R02)))
        rm12 = r2cv*(1-sqrt(abs(r2cv-scaled_R102)))
        avgrmcv = (rm2+rm12)/2
        deltarmcv = abs(rm2-rm12)
        F = (np.shape(self.X)[0]-nLV-1)*R2_value/(nLV*(1-R2_value))
        param = [press_value, R2_value, RMSEC_value, rcal_value, avgrmcal, deltarmcal, presscv_value, Q2_value,
                 RMSECV_value, rcv_value, avgrmcv, deltarmcv, F, nLV]
        dfcv = pd.DataFrame(index=["PRESS", "R2", "RMSEC", "rcal", "avgRmcal", "deltaRmcal", "PRESSCV", "Q2",
                                   "RMSECV", "rcv", "avgRmcv", "deltaRmcv", "F", "nLV"],
                            data=np.array(param))
        return dfcv

    def saveParameters(self, fileName, nLV=None) -> None:
        """Save `filename` containing the calculated parameters for CrossValidation.

        Args:
            fileName (str, path, file-like, `io`): The filename to be saved.
            nLV (int, optional): Number of Latent Variables. Defaults to None.
        """
        df = self.returnParameters(nLV)
        df.to_csv(fileName, sep=',', header=False)



"""Implements Typing constraints to config data passed to some modules. For instance, GA and OPS both accepts a configuration Dictionary that follow these rules.
    """

try:
    from typing import TypedDict  # Python >= 3.8
except ImportError:
    from typing_extensions import TypedDict  # Python < 3.8
from typing import Union


class ConfigGAInterface(TypedDict):
    XMatrix: str
    yvector: str
    varcut: float
    corrcut: float
    max_latent_model: Union[int, None]
    min_vars_model: int
    max_vars_model: int
    population_size: int
    migration_rate: float
    crossover_rate: float
    mutation_rate: float
    generations: int
    yrand: float
    lno: float
    output_matrix: str
    output_cv: str
    output_q2: str
    output_selected: str
    autoscale: bool
    lj_transform: bool
    autocorrcut: float


class ConfigOPSInterface(TypedDict):
    XMatrix: str
    yvector: str
    varcut: float
    corrcut: float
    latent_vars_ops: int
    latent_vars_model: int
    ops_window: int
    ops_increment: int
    vars_percentage: float
    models_to_save: int
    yrand: float
    lno: float
    output_matrix: str
    output_cv: str
    output_models: str
    lj_transform: bool
    autoscale: bool
    autocorrcut: float
    ops_type: str



import numpy as np
from sklearn.preprocessing import scale


def normalize(v):
    (m, n) = np.shape(v)
    fn = np.zeros(n)
    vn = np.zeros((m, n))
    for i in range(n):
        fn[i] = np.linalg.norm(v[:, i])
        vn[:, i] = v[:, i]/fn[i]
    return vn, fn


class PLSBidiag(object):
    """[PLS Bidiagonal](https://onlinelibrary.wiley.com/doi/abs/10.1002/cem.1309) method.
    """

    def __init__(self, n_components=1, scale=True):
        self.nLV = n_components
        self.scale = scale

    def bdiag(self, a, y):
        nLV = self.nLV
        (nrow, ncol) = np.shape(a)
        v = np.zeros((ncol, nLV))
        w = np.zeros(nLV)
        u = np.zeros((nrow, nLV))
        rv1 = np.zeros(nLV)
        for i in range(nLV):
            vi = normalize(np.dot(np.transpose(a), y))[0]
            ui = np.dot(a, vi)
            if i == 0:
                ui, nor = normalize(ui)
            else:
                temp = np.dot(np.transpose(u), ui)
                rv1[i-1] = temp[i-1]
                ui, nor = normalize(ui - np.dot(u, temp))
            w[i] = nor[0]
            v[:, i] = vi[:, 0]
            u[:, i] = ui[:, 0]
            if i < nLV-1:
                y = y - np.dot(ui, np.dot(np.transpose(ui), y))
        return u, w, rv1, v

    def fit(self, X, y):
        if self.scale:
            a = scale(X)
            yt = scale(y)
        else:
            a = X
            yt = yt
        U, w, rv1, V = self.bdiag(a, yt)
        S = np.diag(w) + np.diag(rv1[:-1], 1)
        m, n = np.shape(X)
        nLV = self.nLV
        B = np.zeros((n, nLV))
        for i in range(nLV):
            c = np.dot(np.transpose(U[:, 0:(i+1)]), yt)
            q = np.linalg.solve(S[0:i+1, 0:i+1], c)
            B[:, i] = np.dot(V[:, 0:(i+1)], q)[:, 0]
        self.Bproc = B
        self.B, self.indT = self.reprocessB(X, y)
        self.U = U
        self.S = S
        self.V = V

    def reprocessB(self, X, y):
        stdy = np.std(y)
        meany = np.mean(y)
        stdX = np.std(X, 0)
        meanX = np.mean(X, 0)
        m, n = np.shape(X)
        nLV = self.nLV
        B = np.zeros((n, nLV))
        indT = np.zeros(nLV)
        Bproc = self.Bproc
        for i in range(nLV):
            B[:, i] = Bproc[:, i]*stdy
            B[:, i] = np.divide(B[:, i], np.transpose(stdX))
            t = np.divide(Bproc[:, i], np.transpose(stdX))
            t = t*stdy
            t = np.dot(meanX, t)
            t = -t + meany
            indT[i] = t
        return B, indT

    def predict(self, X):
        B = self.B
        indT = self.indT
        yp = np.dot(X, B)
        for i in range(len(indT)):
            yp[:, i] = yp[:, i] + indT[i]
        return yp



""" **DEPRECATED.**

Throw an error when imported. Use `qsarmodelingpy.external_validation` instead."""
import logging

logging.error(
    "`qsarmodelingpy.validacao_externa` was removed. Use `qsarmodelingpy.external_validation` instead.")

###############################
###############################
###############################
###############################
###############################
###############################
# from sklearn.model_selection import ShuffleSplit
# from sklearn.cross_decomposition import PLSRegression
# from qsarmodelingpy.cross_validation_class import CrossValidation
# from qsarmodelingpy.calculate_parameters import ssy, calcPress, calcR, calcR2, calcMAE, calcRMSE
# import numpy as np
# import pandas as pd
# import math
# import sys
# import json
# import os


# class ExternalValidation(object):
#     def __init__(self, X, y, nLV=None):
#         super(ExternalValidation, self).__init__()
#         self.X = X
#         self.y = y
#         self.nLV = nLV

#     def validateExtVal(self, train, test, nLV=None):
#         X = self.X
#         y = self.y
#         nLV = self.nLV
#         if nLV == None:
#             cv = CrossValidation(X[train, :], y[train])
#             nLV = np.argmax(cv.Q2())+1
#         pls = PLSRegression(n_components=nLV)
#         pls.fit(X[train, :], y[train])
#         yp = pls.predict(X[test])
#         Q2F1 = calcR2(y[test], yp, np.mean(y[train]))
#         Q2F2 = calcR2(y[test], yp)
#         # Q2F3 = 1-((ytest-ypred).T.dot(ytest-ypred)/len(ytest))/((ytrain-ytrain.mean()).T.dot(ytrain-ytrain.mean())/len(ytrain))
#         MAE = calcMAE(y[test], yp)
#         residuals = np.reshape(yp-y[test], len(test))
#         sd = np.std(abs(residuals), ddof=1)
#         higher_residuals = np.argsort(-abs(residuals)
#                                       )[:math.ceil(0.05*len(test))]
#         higher_residuals = [test[i] for i in higher_residuals]
#         new_test = [i for i in test if i not in higher_residuals]
#         yp = pls.predict(X[new_test])
#         Q2F195 = calcR2(y[new_test], yp, np.mean(y[train]))
#         Q2F295 = calcR2(y[new_test], yp)
#         MAE95 = calcMAE(y[new_test], yp)
#         residuals95 = np.reshape(abs(yp-y[new_test]), len(new_test))
#         sd95 = np.std(residuals95, ddof=1)
#         tr = max(y[train]) - min(y[train])
#         NE = [i for i in residuals if i < 0]
#         PE = [i for i in residuals if i > 0]
#         nNE = len(NE)
#         nPE = len(PE)
#         mNE = abs(np.mean(NE))
#         mPE = np.mean(PE)
#         if (nNE == 0) or (nPE == 0) or (nNE/nPE > 5) or (nPE/nNE > 5):
#             return False
#         if (MAE95 <= 0.1*tr) and (MAE95 + 3*sd95 <= 0.2*tr):
#             if Q2F1 > 0.5 and Q2F2 > 0.5:
#                 return True
#             else:
#                 return False
#         else:
#             return False

#     def searchValidExtVal(self, nLV=None, n_test=10, n_splits=100):
#         X = self.X
#         y = self.y
#         nLV = self.nLV
#         if nLV == None:
#             cv = CrossValidation(X, y)
#             nLV = np.argmax(cv.Q2())+1
#         N_SPLITS = n_splits
#         TEST_SIZE = n_test/len(y)
#         ss = ShuffleSplit(n_splits=N_SPLITS, test_size=TEST_SIZE)
#         pls = PLSRegression(n_components=nLV)
#         # yp = np.zeros((self.n_test,N_SPLITS))
#         # R2 = np.zeros(N_SPLITS)
#         R2 = np.zeros(0)
#         Q2 = np.zeros(0)
#         i = 0
#         tests_sets = []
#         for train, test in ss.split(X):
#             # teste com conjunto teste fixo para ver se parametros estão sendo calculados corretamente
#             # test = [23, 47, 6, 25, 40, 13, 48, 30, 43, 19]
#             # train = [i for i in range(len(y)) if i not in test]
#             cv = CrossValidation(X[train, :], y[train])
#             nLV = np.argmax(cv.Q2())+1
#             pls = PLSRegression(n_components=nLV)
#             pls.fit(X[train, :], y[train])
#             yp = pls.predict(X[test])
#             Q2F1 = calcR2(y[test], yp, np.mean(y[train]))
#             Q2F2 = calcR2(y[test], yp)
#             MAE = calcMAE(y[test], yp)
#             # print(Q2F1)
#             # print(Q2F2)
#             # print(MAE)
#             residuals = np.reshape(yp-y[test], len(test))
#             sd = np.std(abs(residuals), ddof=1)
#             # print(sd)
#             higher_residuals = np.argsort(-abs(residuals)
#                                           )[:math.ceil(0.05*n_test)]
#             higher_residuals = [test[i] for i in higher_residuals]
#             new_test = [i for i in test if i not in higher_residuals]
#             yp = pls.predict(X[new_test])
#             Q2F195 = calcR2(y[new_test], yp, np.mean(y[train]))
#             Q2F295 = calcR2(y[new_test], yp)
#             MAE95 = calcMAE(y[new_test], yp)
#             residuals95 = np.reshape(abs(yp-y[new_test]), len(new_test))
#             sd95 = np.std(residuals95, ddof=1)
#             # print(Q2F195)
#             # print(Q2F295)
#             # print(MAE95)
#             # print(sd95)
#             # print(MAE95 + 3*sd95)
#             tr = max(y[train]) - min(y[train])
#             # print(tr)
#             # print(0.1*tr)
#             # print(0.15*tr)
#             # print(0.2*tr)
#             # print(0.25*tr)
#             NE = [i for i in residuals if i < 0]
#             PE = [i for i in residuals if i > 0]
#             nNE = len(NE)
#             nPE = len(PE)
#             mNE = abs(np.mean(NE))
#             mPE = np.mean(PE)
#             # print(nNE/nPE)
#             # print(nPE/nNE)
#             # print(mNE/mPE)
#             # print(mPE/mNE)
#             if (nNE == 0) or (nPE == 0) or (nNE/nPE > 5) or (nPE/nNE > 5):
#                 print("Warning")
#             if (MAE95 <= 0.1*tr) and (MAE95 + 3*sd95 <= 0.2*tr):
#                 print("good")
#                 # print(MAE95)
#                 # print(MAE95 + 3*sd95)
#                 # print(0.1*tr)
#                 # print(0.15*tr)
#                 # print(0.2*tr)
#                 # print(0.25*tr)
#                 # print(test)
#                 tests_sets.append(test)
#                 R2 = np.r_[R2, Q2F2]
#                 Q2 = np.r_[Q2, max(cv.Q2())]
#             else:
#                 if (MAE95 > 0.15*tr) or (MAE95 + 3*sd95 > 0.25*tr):
#                     print("bad")
#                 else:
#                     print("moderate")

#         ind = np.argsort(-R2)
#         self.R2 = R2[ind]
#         self.Q2 = Q2[ind]
#         self.tests_sets = [tests_sets[i] for i in ind]

#     def extVal(self, train, test, nLV=None):
#         Xtest = self.X[test]
#         Xtrain = self.X[train]
#         ytest = self.y[test]
#         ytrain = self.y[train]
#         nLV = self.nLV
#         if nLV == None:
#             cv = CrossValidation(Xtrain, ytrain)
#             nLV = np.argmax(cv.Q2())+1
#         pls = PLSRegression(n_components=nLV)
#         pls.fit(Xtrain, ytrain)
#         ypred = pls.predict(Xtest)
#         self.ypred = ypred
#         Q2F1 = calcR2(ytest, ypred, np.mean(ytrain))
#         Q2F2 = calcR2(ytest, ypred)
#         Q2F3 = 1-((ytest-ypred).T.dot(ytest-ypred)/len(ytest)) / \
#             ((ytrain-ytrain.mean()).T.dot(ytrain-ytrain.mean())/len(ytrain))
#         residuals = np.reshape(ypred-ytest, len(test))
#         higher_residuals = np.argsort(-abs(residuals)
#                                       )[:math.ceil(0.05*len(test))]
#         higher_residuals = [test[i] for i in higher_residuals]
#         new_test = [i for i in test if i not in higher_residuals]
#         yp = pls.predict(self.X[new_test])
#         Q2F195 = calcR2(self.y[new_test], yp, np.mean(ytrain))
#         Q2F295 = calcR2(self.y[new_test], yp)
#         Q2F395 = 1-((self.y[new_test]-yp).T.dot(self.y[new_test]-yp)/len(yp)) / \
#             ((ytrain-ytrain.mean()).T.dot(ytrain-ytrain.mean())/len(ytrain))
#         MAE95 = calcMAE(self.y[new_test], yp)
#         k = sum(ytest*ypred)/sum(ypred**2)
#         k1 = sum(ytest*ypred)/sum(ytest**2)
#         yr0 = k*ypred
#         y1r0 = k1*ytest
#         R02 = calcR2(ytest, yr0)
#         R102 = calcR2(ypred, y1r0)
#         dif = abs(R02-R102)
#         AREpred = sum(abs(ytest-ypred)/ytest)*100/len(ytest)
#         RMSEP = calcRMSE(ytest, ypred)
#         scaledy_test = (ytest-min(ytest))/(max(ytest)-min(ytest))
#         scaledyp = (ypred-min(ytest))/(max(ytest)-min(ytest))
#         scaled_k = sum(scaledy_test*scaledyp)/sum(scaledyp**2)
#         scaled_k1 = sum(scaledy_test*scaledyp)/sum(scaledy_test**2)
#         scaled_yr0 = scaled_k*scaledyp
#         scaled_y1r0 = scaled_k1*scaledy_test
#         scaled_R02 = calcR2(scaledy_test, scaled_yr0)
#         scaled_R102 = calcR2(scaledyp, scaled_y1r0)
#         r2 = (calcR(np.reshape(scaledy_test, len(scaledy_test)),
#                     np.reshape(scaledyp, len(scaledy_test))))**2
#         rm2 = r2*(1-math.sqrt(r2-scaled_R02))
#         rm12 = r2*(1-math.sqrt(r2-scaled_R102))
#         avgrm = (rm2+rm12)/2
#         deltarm = abs(rm2-rm12)
#         MAE = calcMAE(ytest, ypred)
#         test_set = str([i+1 for i in test])
#         # param = [Q2F1,Q2F2,k,k1,R02,R102,dif,AREpred,RMSEP,str(test)]
#         # dfPred = pd.DataFrame(index = ["Q2F1","Q2F2","k","k'","R0","R0'","dif","AREpred","RMSEP","test_set"],
#         #     data=np.array([Q2F1[0],Q2F2[0],k[0],k1[0],R02[0],R102[0],dif[0],AREpred[0],RMSEP,str(test)]))
#         # param = [Q2F1,Q2F2,k,k1,R02,R102,dif,AREpred,RMSEP,avgrm,deltarm,str(test)]
#         dfPred = pd.DataFrame(index=["Q2F1", "Q2F2", "Q2F3", "Q2F195", "Q2F295", "Q2F395",
#                                      "k", "k'", "R0",
#                                      "R0'", "dif", "AREpred", "RMSEP", "avgrm",
#                                      "deltarm", "MAE", "MAE95", "test_set", "nLV"],
#                               data=np.array([Q2F1[0], Q2F2[0], Q2F3[0][0], Q2F195[0], Q2F295[0], Q2F395[0][0],
#                                              k[0], k1[0], R02[0], R102[0], dif[0], AREpred[0], RMSEP, avgrm,
#                                              deltarm, MAE, MAE95, test_set, nLV]))
#         return dfPred

#     def saveExtVal(self, train, test, fileName):
#         dfPred = self.extVal(train, test, self.nLV)
#         dfPred.to_csv(fileName, sep=',', header=False)

#     def runValidExtVal(self, directory, n_models=1):
#         print("Entrou na funcao de testar validacoes")
#         print(len(self.tests_sets))
#         for i in range(min(n_models, len(self.tests_sets))):
#             test = self.tests_sets[i]
#             train = [j for j in range(len(self.y)) if j not in test]
#             dfPred = self.extVal(train, test)
#             dfPred.to_csv(directory+"test"+str(i+1) +
#                           ".csv", sep=',', header=False)

#     def searchValidExtVal2(self, directory, nLV=None, n_test=10, n_splits=100):
#         X = self.X
#         y = self.y
#         nLV = self.nLV
#         if nLV == None:
#             cv = CrossValidation(X, y)
#             nLV = np.argmax(cv.Q2())+1
#         N_SPLITS = n_splits
#         TEST_SIZE = n_test/len(y)
#         ss = ShuffleSplit(n_splits=N_SPLITS, test_size=TEST_SIZE)
#         pls = PLSRegression(n_components=nLV)
#         # yp = np.zeros((self.n_test,N_SPLITS))
#         # R2 = np.zeros(N_SPLITS)
#         # R2 = np.zeros(0)
#         # Q2 = np.zeros(0)
#         i = 0
#         # tests_sets = []
#         dfExt = pd.DataFrame()
#         dfCv = pd.DataFrame()
#         for train, test in ss.split(X):
#             if self.validateExtVal(train, test, nLV):
#                 # print("Achou um conjunto teste valido")
#                 # self.saveExtVal(train,test,os.path.join(directory,"test"+str(i+1)+".csv"))
#                 dfExt[i] = self.extVal(train, test, nLV).iloc[:, 0]
#                 cv = CrossValidation(
#                     X[train, :], y[train], nLVMax=nLV, scale=True)
#                 dfCv[i] = cv.returnParameters(nLV).iloc[:, 0]
#                 # cv.saveParameters(os.path.join(directory,"test_cv"+str(i+1)+".csv"))
#                 i = i + 1
#         return dfExt, dfCv


# # R2 = R2[ind]
# # yp = yp[:,ind]

# # ytosave = np.zeros((test_size,0))
# # R2tosave = np.zeros(0)
# # teststosave = []

# # i = 0
# # count = 0
# # while R2[i] > 0.5:
# #     residue = True
# #     for j in range(test_size):
# #         if abs(yp[j,i] - y[tests_sets[ind[i]][j]][0]) > 0.5:
# #             residue = False
# #     if residue:
# #         count += 1
# #         ytosave = np.c_[ytosave,yp[:,i]]
# #         R2tosave = np.r_[R2tosave,R2[i]]
# #         teststosave.append(tests_sets[ind[i]])
# #         dfPred = pd.DataFrame(data=np.c_[y[tests_sets[ind[i]]],yp[:,i]],columns=["Yreal","Ypred"])
# #         dfPred.to_csv("test"+str(count)+".csv", sep =',', index=False)
# #     i += 1


# # with open(out_directory+"/R2.json", "w") as f:
# #     f.write(json.dumps(R2tosave.tolist()))

# # with open(out_directory+"/yp.json", "w") as f:
# #     f.write(json.dumps(ytosave.tolist()))

# # with open(out_directory+"/tests.json", "w") as f:
# #     f.write(json.dumps(teststosave))

# # df = pd.DataFrame(ytosave)
# # df2 = pd.DataFrame(R2tosave)
# # df2 = df2.transpose()
# # str_tests = [str(teststosave[i]) for i in range(len(teststosave))]
# # df3 = pd.DataFrame(str_tests)
# # df3 = df3.transpose()
# # frames = [df,df2,df3]
# # result = pd.concat(frames)
# # print(result.shape)
# # # df.to_csv('yp.csv', sep =';')
# # # df2.to_csv('R2.csv', sep =';')
# # df3.to_csv('tests.csv', sep =';')
# # result.to_csv('result.csv', sep =';')
